"""OnboardingTask Django model.

(c) 2020 Network To Code
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import socket

import netaddr
from netaddr.core import AddrFormatError

from .exceptions import OnboardException


def onboarding_task_fqdn_to_ip(ot):
    """Method to assure OT has FQDN resolved to IP address and rewritten into OT.

    If it is a DNS name, attempt to resolve the DNS address and assign the IP address to the
    name.

    Returns:
        None

    Raises:
      OnboardException("fail-general"):
        When a prefix was entered for an IP address
      OnboardException("fail-dns"):
        When a Name lookup via DNS fails to resolve an IP address
    """
    try:
        # If successful, this is an IP address and can pass
        netaddr.IPAddress(ot.ip_address)
    # Raise an Exception for Prefix values
    except ValueError:
        raise OnboardException(reason="fail-general", message=f"ERROR appears a prefix was entered: {ot.ip_address}")
    # An AddrFormatError exception means that there is not an IP address in the field, and should continue on
    except AddrFormatError:
        try:
            # Perform DNS Lookup
            ot.ip_address = socket.gethostbyname(ot.ip_address)
            ot.save()
        except socket.gaierror:
            # DNS Lookup has failed, Raise an exception for unable to complete DNS lookup
            raise OnboardException(reason="fail-dns", message=f"ERROR failed to complete DNS lookup: {ot.ip_address}")



"""NetBox Keeper.

(c) 2020 Network To Code
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import re

from django.conf import settings
from django.utils.text import slugify
from dcim.models import Manufacturer, Device, Interface, DeviceType, DeviceRole
from dcim.models import Platform
from dcim.models import Site
from ipam.models import IPAddress

from .constants import NETMIKO_TO_NAPALM_STATIC
from .exceptions import OnboardException

logger = logging.getLogger("rq.worker")

PLUGIN_SETTINGS = settings.PLUGINS_CONFIG["netbox_onboarding"]


def object_match(obj, search_array):
    """Used to search models for multiple criteria.

    Inputs:
        obj:            The model used for searching.
        search_array:   Nested dictionaries used to search models. First criteria will be used
                        for strict searching. Loose searching will loop through the search_array
                        until it finds a match. Example below.
                        [
                            {"slug__iexact": 'switch1'},
                            {"model__iexact": 'Cisco'}
                        ]
    """
    try:
        result = obj.objects.get(**search_array[0])
        return result
    except obj.DoesNotExist:
        if PLUGIN_SETTINGS["object_match_strategy"] == "loose":
            for search_array_element in search_array[1:]:
                try:
                    result = obj.objects.get(**search_array_element)
                    return result
                except obj.DoesNotExist:
                    pass
                except obj.MultipleObjectsReturned:
                    raise OnboardException(
                        reason="fail-general",
                        message=f"ERROR multiple objects found in {str(obj)} searching on {str(search_array_element)})",
                    )
        raise
    except obj.MultipleObjectsReturned:
        raise OnboardException(
            reason="fail-general",
            message=f"ERROR multiple objects found in {str(obj)} searching on {str(search_array_element)})",
        )


class NetboxKeeper:
    """Used to manage the information relating to the network device within the NetBox server."""

    def __init__(  # pylint: disable=R0913,R0914
        self,
        netdev_hostname,
        netdev_nb_role_slug,
        netdev_vendor,
        netdev_nb_site_slug,
        netdev_nb_device_type_slug=None,
        netdev_model=None,
        netdev_nb_role_color=None,
        netdev_mgmt_ip_address=None,
        netdev_nb_platform_slug=None,
        netdev_serial_number=None,
        netdev_mgmt_ifname=None,
        netdev_mgmt_pflen=None,
        netdev_netmiko_device_type=None,
        onboarding_class=None,
        driver_addon_result=None,
    ):
        """Create an instance and initialize the managed attributes that are used throughout the onboard processing.

        Args:
            netdev_hostname (str): NetBox's device name
            netdev_nb_role_slug (str): NetBox's device role slug
            netdev_vendor (str): Device's vendor name
            netdev_nb_site_slug (str): Device site's slug
            netdev_nb_device_type_slug (str): Device type's slug
            netdev_model (str): Device's model
            netdev_nb_role_color (str): NetBox device's role color
            netdev_mgmt_ip_address (str): IPv4 Address of a device
            netdev_nb_platform_slug (str): NetBox device's platform slug
            netdev_serial_number (str): Device's serial number
            netdev_mgmt_ifname (str): Device's management interface name
            netdev_mgmt_pflen (str): Device's management IP prefix-len
            netdev_netmiko_device_type (str): Device's Netmiko device type
            onboarding_class (Object): Onboarding Class (future use)
            driver_addon_result (Any): Attached extended result (future use)
        """
        self.netdev_mgmt_ip_address = netdev_mgmt_ip_address
        self.netdev_nb_site_slug = netdev_nb_site_slug
        self.netdev_nb_device_type_slug = netdev_nb_device_type_slug
        self.netdev_nb_role_slug = netdev_nb_role_slug
        self.netdev_nb_role_color = netdev_nb_role_color
        self.netdev_nb_platform_slug = netdev_nb_platform_slug

        self.netdev_hostname = netdev_hostname
        self.netdev_vendor = netdev_vendor
        self.netdev_model = netdev_model
        self.netdev_serial_number = netdev_serial_number
        self.netdev_mgmt_ifname = netdev_mgmt_ifname
        self.netdev_mgmt_pflen = netdev_mgmt_pflen
        self.netdev_netmiko_device_type = netdev_netmiko_device_type

        self.onboarding_class = onboarding_class
        self.driver_addon_result = driver_addon_result

        # these attributes are netbox model instances as discovered/created
        # through the course of processing.
        self.nb_site = None
        self.nb_manufacturer = None
        self.nb_device_type = None
        self.nb_device_role = None
        self.nb_platform = None

        self.device = None
        self.onboarded_device = None
        self.nb_mgmt_ifname = None
        self.nb_primary_ip = None

    def ensure_onboarded_device(self):
        """Lookup if the device already exists in the NetBox.

        Lookup is performed by querying for the IP address of the onboarded device.
        If the device with a given IP is already in NetBox, its attributes including name could be updated
        """
        try:
            if self.netdev_mgmt_ip_address:
                self.onboarded_device = Device.objects.get(primary_ip4__address__net_host=self.netdev_mgmt_ip_address)
        except Device.DoesNotExist:
            logger.info(
                "Could not find existing NetBox device for requested primary IP address (%s)",
                self.netdev_mgmt_ip_address,
            )
        except Device.MultipleObjectsReturned:
            raise OnboardException(
                reason="fail-general",
                message=f"ERROR multiple devices using same IP in NetBox: {self.netdev_mgmt_ip_address}",
            )

    def ensure_device_site(self):
        """Ensure device's site."""
        try:
            self.nb_site = Site.objects.get(slug=self.netdev_nb_site_slug)
        except Site.DoesNotExist:
            raise OnboardException(reason="fail-config", message=f"Site not found: {self.netdev_nb_site_slug}")

    def ensure_device_manufacturer(
        self,
        create_manufacturer=PLUGIN_SETTINGS["create_manufacturer_if_missing"],
        skip_manufacturer_on_update=PLUGIN_SETTINGS["skip_manufacturer_on_update"],
    ):
        """Ensure device's manufacturer."""
        # Support to skip manufacturer updates for existing devices
        if self.onboarded_device and skip_manufacturer_on_update:
            self.nb_manufacturer = self.onboarded_device.device_type.manufacturer

            return

        # First ensure that the vendor, as extracted from the network device exists
        # in NetBox.  We need the ID for this vendor when ensuring the DeviceType
        # instance.

        nb_manufacturer_slug = slugify(self.netdev_vendor)

        try:
            search_array = [{"slug__iexact": nb_manufacturer_slug}]
            self.nb_manufacturer = object_match(Manufacturer, search_array)
        except Manufacturer.DoesNotExist:
            if create_manufacturer:
                self.nb_manufacturer = Manufacturer.objects.create(name=self.netdev_vendor, slug=nb_manufacturer_slug)
            else:
                raise OnboardException(
                    reason="fail-config", message=f"ERROR manufacturer not found: {self.netdev_vendor}"
                )

    def ensure_device_type(
        self,
        create_device_type=PLUGIN_SETTINGS["create_device_type_if_missing"],
        skip_device_type_on_update=PLUGIN_SETTINGS["skip_device_type_on_update"],
    ):
        """Ensure the Device Type (slug) exists in NetBox associated to the netdev "model" and "vendor" (manufacturer).

        Args:
          create_device_type (bool): Flag to indicate if we need to create the device_type, if not already present
          skip_device_type_on_update (bool): Flag to indicate if we skip device type updates for existing devices
        Raises:
          OnboardException('fail-config'):
            When the device vendor value does not exist as a Manufacturer in
            NetBox.

          OnboardException('fail-config'):
            When the device-type exists by slug, but is assigned to a different
            manufacturer.  This should *not* happen, but guard-rail checking
            regardless in case two vendors have the same model name.
        """
        # Support to skip device type updates for existing devices
        if self.onboarded_device and skip_device_type_on_update:
            self.nb_device_type = self.onboarded_device.device_type

            return

        # Now see if the device type (slug) already exists,
        #  if so check to make sure that it is not assigned as a different manufacturer
        # if it doesn't exist, create it if the flag 'create_device_type_if_missing' is defined

        slug = self.netdev_model
        if self.netdev_model and re.search(r"[^a-zA-Z0-9\-_]+", slug):
            logger.warning("device model is not sluggable: %s", slug)
            self.netdev_model = slug.replace(" ", "-")
            logger.warning("device model is now: %s", self.netdev_model)

        # Use declared device type or auto-discovered model
        nb_device_type_text = self.netdev_nb_device_type_slug or self.netdev_model

        if not nb_device_type_text:
            raise OnboardException(reason="fail-config", message="ERROR device type not found")

        nb_device_type_slug = slugify(nb_device_type_text)

        try:
            search_array = [
                {"slug__iexact": nb_device_type_slug},
                {"model__iexact": self.netdev_model},
                {"part_number__iexact": self.netdev_model},
            ]

            self.nb_device_type = object_match(DeviceType, search_array)

            if self.nb_device_type.manufacturer.id != self.nb_manufacturer.id:
                raise OnboardException(
                    reason="fail-config",
                    message=f"ERROR device type {self.netdev_model} " f"already exists for vendor {self.netdev_vendor}",
                )

        except DeviceType.DoesNotExist:
            if create_device_type:
                logger.info("CREATE: device-type: %s", self.netdev_model)
                self.nb_device_type = DeviceType.objects.create(
                    slug=nb_device_type_slug, model=nb_device_type_slug.upper(), manufacturer=self.nb_manufacturer,
                )
            else:
                raise OnboardException(
                    reason="fail-config", message=f"ERROR device type not found: {self.netdev_model}"
                )

    def ensure_device_role(
        self, create_device_role=PLUGIN_SETTINGS["create_device_role_if_missing"],
    ):
        """Ensure that the device role is defined / exist in NetBox or create it if it doesn't exist.

        Args:
          create_device_role (bool) :Flag to indicate if we need to create the device_role, if not already present
        Raises:
          OnboardException('fail-config'):
            When the device role value does not exist
            NetBox.
        """
        try:
            self.nb_device_role = DeviceRole.objects.get(slug=self.netdev_nb_role_slug)
        except DeviceRole.DoesNotExist:
            if create_device_role:
                self.nb_device_role = DeviceRole.objects.create(
                    name=self.netdev_nb_role_slug,
                    slug=self.netdev_nb_role_slug,
                    color=self.netdev_nb_role_color,
                    vm_role=False,
                )
            else:
                raise OnboardException(
                    reason="fail-config", message=f"ERROR device role not found: {self.netdev_nb_role_slug}"
                )

    def ensure_device_platform(self, create_platform_if_missing=PLUGIN_SETTINGS["create_platform_if_missing"]):
        """Get platform object from NetBox filtered by platform_slug.

        Args:
            platform_slug (string): slug of a platform object present in NetBox, object will be created if not present
            and create_platform_if_missing is enabled

        Return:
            dcim.models.Platform object

        Raises:
            OnboardException

        Lookup is performed based on the object's slug field (not the name field)
        """
        try:
            self.netdev_nb_platform_slug = (
                self.netdev_nb_platform_slug
                or PLUGIN_SETTINGS["platform_map"].get(self.netdev_netmiko_device_type)
                or self.netdev_netmiko_device_type
            )

            if not self.netdev_nb_platform_slug:
                raise OnboardException(
                    reason="fail-config", message=f"ERROR device platform not found: {self.netdev_hostname}"
                )

            self.nb_platform = Platform.objects.get(slug=self.netdev_nb_platform_slug)

            logger.info("PLATFORM: found in NetBox %s", self.netdev_nb_platform_slug)

        except Platform.DoesNotExist:
            if create_platform_if_missing:
                platform_to_napalm_netbox = {
                    platform.slug: platform.napalm_driver
                    for platform in Platform.objects.all()
                    if platform.napalm_driver
                }

                # Update Constants if Napalm driver is defined for NetBox Platform
                netmiko_to_napalm = {**NETMIKO_TO_NAPALM_STATIC, **platform_to_napalm_netbox}

                self.nb_platform = Platform.objects.create(
                    name=self.netdev_nb_platform_slug,
                    slug=self.netdev_nb_platform_slug,
                    napalm_driver=netmiko_to_napalm[self.netdev_netmiko_device_type],
                )
            else:
                raise OnboardException(
                    reason="fail-general", message=f"ERROR platform not found in NetBox: {self.netdev_nb_platform_slug}"
                )

    def ensure_device_instance(self, default_status=PLUGIN_SETTINGS["default_device_status"]):
        """Ensure that the device instance exists in NetBox and is assigned the provided device role or DEFAULT_ROLE.

        Args:
          default_status (str) : status assigned to a new device by default.
        """
        if self.onboarded_device:
            # Construct lookup arguments if onboarded device already exists in NetBox

            logger.info(
                "Found existing NetBox device (%s) for requested primary IP address (%s)",
                self.onboarded_device.name,
                self.netdev_mgmt_ip_address,
            )
            lookup_args = {
                "pk": self.onboarded_device.pk,
                "defaults": dict(
                    name=self.netdev_hostname,
                    device_type=self.nb_device_type,
                    device_role=self.nb_device_role,
                    platform=self.nb_platform,
                    site=self.nb_site,
                    serial=self.netdev_serial_number,
                    # status= field is not updated in case of already existing devices to prevent changes
                ),
            }
        else:
            # Construct lookup arguments if onboarded device does not exist in NetBox

            lookup_args = {
                "name": self.netdev_hostname,
                "defaults": dict(
                    device_type=self.nb_device_type,
                    device_role=self.nb_device_role,
                    platform=self.nb_platform,
                    site=self.nb_site,
                    serial=self.netdev_serial_number,
                    # status= defined only for new devices, no update for existing should occur
                    status=default_status,
                ),
            }

        try:
            self.device, created = Device.objects.update_or_create(**lookup_args)

            if created:
                logger.info("CREATED device: %s", self.netdev_hostname)
            else:
                logger.info("GOT/UPDATED device: %s", self.netdev_hostname)

        except Device.MultipleObjectsReturned:
            raise OnboardException(
                reason="fail-general",
                message=f"ERROR multiple devices using same name in NetBox: {self.netdev_hostname}",
            )

    def ensure_interface(self):
        """Ensures that the interface associated with the mgmt_ipaddr exists and is assigned to the device."""
        if self.netdev_mgmt_ifname:
            self.nb_mgmt_ifname, _ = Interface.objects.get_or_create(name=self.netdev_mgmt_ifname, device=self.device)

    def ensure_primary_ip(self):
        """Ensure mgmt_ipaddr exists in IPAM, has the device interface, and is assigned as the primary IP address."""
        # see if the primary IP address exists in IPAM
        if self.netdev_mgmt_ip_address and self.netdev_mgmt_pflen:
            self.nb_primary_ip, created = IPAddress.objects.get_or_create(
                address=f"{self.netdev_mgmt_ip_address}/{self.netdev_mgmt_pflen}"
            )

            if created or not self.nb_primary_ip in self.nb_mgmt_ifname.ip_addresses.all():
                logger.info("ASSIGN: IP address %s to %s", self.nb_primary_ip.address, self.nb_mgmt_ifname.name)
                self.nb_mgmt_ifname.ip_addresses.add(self.nb_primary_ip)
                self.nb_mgmt_ifname.save()

            # Ensure the primary IP is assigned to the device
            self.device.primary_ip4 = self.nb_primary_ip
            self.device.save()

    def ensure_device(self):
        """Ensure that the device represented by the DevNetKeeper exists in the NetBox system."""
        self.ensure_onboarded_device()
        self.ensure_device_site()
        self.ensure_device_manufacturer()
        self.ensure_device_type()
        self.ensure_device_role()
        self.ensure_device_platform()
        self.ensure_device_instance()

        if PLUGIN_SETTINGS["create_management_interface_if_missing"]:
            self.ensure_interface()
            self.ensure_primary_ip()



"""Exceptions.

(c) 2020 Network To Code
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class OnboardException(Exception):
    """A failure occurred during the onboarding process.

    The exception includes a reason "slug" as defined below as well as a humanized message.
    """

    REASONS = (
        "fail-config",  # config provided is not valid
        "fail-connect",  # device is unreachable at IP:PORT
        "fail-execute",  # unable to execute device/API command
        "fail-login",  # bad username/password
        "fail-dns",  # failed to get IP address from name resolution
        "fail-general",  # other error
    )

    def __init__(self, reason, message, **kwargs):
        """Exception Init."""
        super(OnboardException, self).__init__(kwargs)
        self.reason = reason
        self.message = message

    def __str__(self):
        """Exception __str__."""
        return f"{self.__class__.__name__}: {self.reason}: {self.message}"



"""Onboarding template content.

(c) 2020 Network To Code
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from extras.plugins import PluginTemplateExtension
from .models import OnboardingDevice


class DeviceContent(PluginTemplateExtension):  # pylint: disable=abstract-method
    """Table to show onboarding details on Device objects."""

    model = "dcim.device"

    def right_page(self):
        """Show table on right side of view."""
        onboarding = OnboardingDevice.objects.filter(device=self.context["object"]).first()

        if not onboarding or not onboarding.enabled:
            return ""

        status = onboarding.status
        last_check_attempt_date = onboarding.last_check_attempt_date
        last_check_successful_date = onboarding.last_check_successful_date
        last_ot = onboarding.last_ot

        return self.render(
            "netbox_onboarding/device_onboarding_table.html",
            extra_context={
                "status": status,
                "last_check_attempt_date": last_check_attempt_date,
                "last_check_successful_date": last_check_successful_date,
                "last_ot": last_ot,
            },
        )


template_extensions = [DeviceContent]



# Generated by Django 2.2.10 on 2020-08-21 11:05

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("netbox_onboarding", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="OnboardingDevice",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False)),
                ("enabled", models.BooleanField(default=True)),
                ("device", models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to="dcim.Device")),
            ],
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("netbox_onboarding", "0002_onboardingdevice"),
    ]

    operations = [
        migrations.AddField(
            model_name="onboardingtask", name="created", field=models.DateField(auto_now_add=True, null=True),
        ),
        migrations.AddField(
            model_name="onboardingtask", name="last_updated", field=models.DateTimeField(auto_now=True, null=True),
        ),
        migrations.AlterModelOptions(name="onboardingtask", options={},),
        migrations.RemoveField(model_name="onboardingtask", name="created_on",),
    ]



"""Unit tests for netbox_onboarding REST API.

(c) 2020 Network To Code
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from django.contrib.auth.models import User  # pylint: disable=imported-auth-user
from django.test import TestCase
from django.urls import reverse
from rest_framework import status
from rest_framework.test import APIClient

from users.models import Token

from dcim.models import Site


from netbox_onboarding.models import OnboardingTask


class OnboardingTaskTestCase(TestCase):
    """Test the OnboardingTask API."""

    def setUp(self):
        """Create a superuser and token for API calls."""
        self.user = User.objects.create(username="testuser", is_superuser=True)
        self.token = Token.objects.create(user=self.user)
        self.client = APIClient()
        self.client.credentials(HTTP_AUTHORIZATION=f"Token {self.token.key}")

        self.base_url_lookup = "plugins-api:netbox_onboarding-api:onboardingtask"

        self.site1 = Site.objects.create(name="USWEST", slug="uswest")

        self.onboarding_task1 = OnboardingTask.objects.create(ip_address="10.10.10.10", site=self.site1)
        self.onboarding_task2 = OnboardingTask.objects.create(ip_address="192.168.1.1", site=self.site1)

    def test_list_onboarding_tasks(self):
        """Verify that OnboardingTasks can be listed."""
        url = reverse(f"{self.base_url_lookup}-list")

        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["count"], 2)

    def test_get_onboarding_task(self):
        """Verify that an Onboardingtask can be retrieved."""
        url = reverse(f"{self.base_url_lookup}-detail", kwargs={"pk": self.onboarding_task1.pk})

        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["ip_address"], self.onboarding_task1.ip_address)
        self.assertEqual(response.data["site"], self.onboarding_task1.site.slug)

    def test_create_task_missing_mandatory_parameters(self):
        """Verify that the only mandatory POST parameters are ip_address and site."""
        url = reverse(f"{self.base_url_lookup}-list")

        response = self.client.post(url, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        # The response tells us which fields are missing from the request
        self.assertIn("ip_address", response.data)
        self.assertIn("site", response.data)
        self.assertEqual(len(response.data), 2, "Only two parameters should be mandatory")

    def test_create_task(self):
        """Verify that an OnboardingTask can be created."""
        url = reverse(f"{self.base_url_lookup}-list")
        data = {"ip_address": "10.10.10.20", "site": self.site1.slug}

        response = self.client.post(url, data, format="json")
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        for key, value in data.items():
            self.assertEqual(response.data[key], value)
        self.assertEqual(response.data["port"], 22)  # default value
        self.assertEqual(response.data["timeout"], 30)  # default value

        onboarding_task = OnboardingTask.objects.get(pk=response.data["id"])
        self.assertEqual(onboarding_task.ip_address, data["ip_address"])
        self.assertEqual(onboarding_task.site, self.site1)

    def test_update_task_forbidden(self):
        """Verify that an OnboardingTask cannot be updated via this API."""
        url = reverse(f"{self.base_url_lookup}-detail", kwargs={"pk": self.onboarding_task1.pk})

        response = self.client.patch(url, {"ip_address": "10.10.10.20"}, format="json")
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)
        self.assertEqual(self.onboarding_task1.ip_address, "10.10.10.10")

        response = self.client.put(url, {"ip_address": "10.10.10.20"}, format="json")
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)
        self.assertEqual(self.onboarding_task1.ip_address, "10.10.10.10")

    def test_delete_task(self):
        """Verify that an OnboardingTask can be deleted."""
        url = reverse(f"{self.base_url_lookup}-detail", kwargs={"pk": self.onboarding_task1.pk})

        response = self.client.delete(url)
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(OnboardingTask.DoesNotExist):
            OnboardingTask.objects.get(pk=self.onboarding_task1.pk)



"""Tables for device onboarding tasks.

(c) 2020 Network To Code
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import django_tables2 as tables
from utilities.tables import BaseTable, ToggleColumn
from .models import OnboardingTask


class OnboardingTaskTable(BaseTable):
    """Table for displaying OnboardingTask instances."""

    pk = ToggleColumn()
    id = tables.LinkColumn()
    site = tables.LinkColumn()
    platform = tables.LinkColumn()
    created_device = tables.LinkColumn()

    class Meta(BaseTable.Meta):  # noqa: D106 "Missing docstring in public nested class"
        model = OnboardingTask
        fields = (
            "pk",
            "id",
            "created",
            "ip_address",
            "site",
            "platform",
            "created_device",
            "status",
            "failed_reason",
            "message",
        )


class OnboardingTaskFeedBulkTable(BaseTable):
    """TODO document me."""

    site = tables.LinkColumn()

    class Meta(BaseTable.Meta):  # noqa: D106 "Missing docstring in public nested class"
        model = OnboardingTask
        fields = (
            "id",
            "created",
            "site",
            "platform",
            "ip_address",
            "port",
            "timeout",
        )



"""Worker code for processing inbound OnboardingTasks.

(c) 2020 Network To Code
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging

from django.core.exceptions import ValidationError
from django_rq import job
from prometheus_client import Summary

from dcim.models import Device

from .choices import OnboardingFailChoices
from .choices import OnboardingStatusChoices
from .exceptions import OnboardException
from .helpers import onboarding_task_fqdn_to_ip
from .metrics import onboardingtask_results_counter
from .models import OnboardingDevice
from .models import OnboardingTask
from .onboard import OnboardingManager

logger = logging.getLogger("rq.worker")


REQUEST_TIME = Summary("onboardingtask_processing_seconds", "Time spent processing onboarding request")


@REQUEST_TIME.time()
@job("default")
def onboard_device(task_id, credentials):  # pylint: disable=too-many-statements
    """Process a single OnboardingTask instance."""
    username = credentials.username
    password = credentials.password
    secret = credentials.secret

    ot = OnboardingTask.objects.get(id=task_id)

    # Rewrite FQDN to IP for Onboarding Task
    onboarding_task_fqdn_to_ip(ot)

    logger.info("START: onboard device")
    onboarded_device = None

    try:
        try:
            if ot.ip_address:
                onboarded_device = Device.objects.get(primary_ip4__address__net_host=ot.ip_address)

            if OnboardingDevice.objects.filter(device=onboarded_device, enabled=False):
                ot.status = OnboardingStatusChoices.STATUS_SKIPPED

                return dict(ok=True)

        except Device.DoesNotExist as exc:
            logger.info("Getting device with IP lookup failed: %s", str(exc))
        except Device.MultipleObjectsReturned as exc:
            logger.info("Getting device with IP lookup failed: %s", str(exc))
            raise OnboardException(
                reason="fail-general", message=f"ERROR Multiple devices exist for IP {ot.ip_address}"
            )
        except ValueError as exc:
            logger.info("Getting device with IP lookup failed: %s", str(exc))
        except ValidationError as exc:
            logger.info("Getting device with IP lookup failed: %s", str(exc))

        ot.status = OnboardingStatusChoices.STATUS_RUNNING
        ot.save()

        onboarding_manager = OnboardingManager(ot=ot, username=username, password=password, secret=secret)

        if onboarding_manager.created_device:
            ot.created_device = onboarding_manager.created_device

        ot.status = OnboardingStatusChoices.STATUS_SUCCEEDED
        ot.save()
        logger.info("FINISH: onboard device")
        onboarding_status = True

    except OnboardException as exc:
        if onboarded_device:
            ot.created_device = onboarded_device

        logger.error("%s", exc)
        ot.status = OnboardingStatusChoices.STATUS_FAILED
        ot.failed_reason = exc.reason
        ot.message = exc.message
        ot.save()
        onboarding_status = False

    except Exception as exc:  # pylint: disable=broad-except
        if onboarded_device:
            ot.created_device = onboarded_device

        logger.error("Onboarding Error - Exception")
        logger.error(str(exc))
        ot.status = OnboardingStatusChoices.STATUS_FAILED
        ot.failed_reason = OnboardingFailChoices.FAIL_GENERAL
        ot.message = str(exc)
        ot.save()
        onboarding_status = False

    onboardingtask_results_counter.labels(status=ot.status).inc()

    return dict(ok=onboarding_status)



"""Onboarding module.

(c) 2020 Network To Code
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from netbox_onboarding.netbox_keeper import NetboxKeeper


class Onboarding:
    """Generic onboarding class."""

    def __init__(self):
        """Init the class."""
        self.created_device = None
        self.credentials = None

    def run(self, onboarding_kwargs):
        """Implement run method."""
        raise NotImplementedError


class StandaloneOnboarding(Onboarding):
    """Standalone onboarding class."""

    def run(self, onboarding_kwargs):
        """Ensure device is created with NetBox Keeper."""
        nb_k = NetboxKeeper(**onboarding_kwargs)
        nb_k.ensure_device()

        self.created_device = nb_k.device



"""NetDev Keeper.

(c) 2020 Network To Code
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import importlib
import logging
import socket

from django.conf import settings
from napalm import get_network_driver
from napalm.base.exceptions import ConnectionException, CommandErrorException
from napalm.base.netmiko_helpers import netmiko_args
from netmiko.ssh_autodetect import SSHDetect
from netmiko.ssh_exception import NetMikoAuthenticationException
from netmiko.ssh_exception import NetMikoTimeoutException
from paramiko.ssh_exception import SSHException

from dcim.models import Platform

from netbox_onboarding.onboarding.onboarding import StandaloneOnboarding
from .constants import NETMIKO_TO_NAPALM_STATIC
from .exceptions import OnboardException

logger = logging.getLogger("rq.worker")

PLUGIN_SETTINGS = settings.PLUGINS_CONFIG["netbox_onboarding"]


def get_mgmt_info(
    hostname,
    ip_ifs,
    default_mgmt_if=PLUGIN_SETTINGS["default_management_interface"],
    default_mgmt_pfxlen=PLUGIN_SETTINGS["default_management_prefix_length"],
):
    """Get the interface name and prefix length for the management interface.

    Locate the interface assigned with the hostname value and retain
    the interface name and IP prefix-length so that we can use it
    when creating the IPAM IP-Address instance.

    Note that in some cases (e.g., NAT) the hostname may differ than
    the interface addresses present on the device. We need to handle this.
    """
    for if_name, if_data in ip_ifs.items():
        for if_addr, if_addr_data in if_data["ipv4"].items():
            if if_addr == hostname:
                return if_name, if_addr_data["prefix_length"]

    return default_mgmt_if, default_mgmt_pfxlen


class NetdevKeeper:
    """Used to maintain information about the network device during the onboarding process."""

    def __init__(  # pylint: disable=R0913
        self,
        hostname,
        port=None,
        timeout=None,
        username=None,
        password=None,
        secret=None,
        napalm_driver=None,
        optional_args=None,
    ):
        """Initialize the network device keeper instance and ensure the required configuration parameters are provided.

        Args:
          hostname (str): IP Address or FQDN of an onboarded device
          port (int): Port used to connect to an onboarded device
          timeout (int): Connection timeout of an onboarded device
          username (str): Device username (if unspecified, NAPALM_USERNAME settings variable will be used)
          password (str): Device password (if unspecified, NAPALM_PASSWORD settings variable will be used)
          secret (str): Device secret password (if unspecified, NAPALM_ARGS["secret"] settings variable will be used)
          napalm_driver (str): Napalm driver name to use to onboard network device

        Raises:
          OnboardException('fail-config'):
            When any required config options are missing.
        """
        # Attributes
        self.hostname = hostname
        self.port = port
        self.timeout = timeout
        self.username = username
        self.password = password
        self.secret = secret
        self.napalm_driver = napalm_driver
        self.optional_args = optional_args

        self.facts = None
        self.ip_ifs = None
        self.netmiko_device_type = None
        self.onboarding_class = StandaloneOnboarding
        self.driver_addon_result = None

        # Enable loading driver extensions
        self.load_driver_extension = True

    def check_reachability(self):
        """Ensure that the device at the mgmt-ipaddr provided is reachable.

        We do this check before attempting other "show" commands so that we know we've got a
        device that can be reached.

        Raises:
          OnboardException('fail-connect'):
            When device unreachable
        """
        logger.info("CHECK: IP %s:%s", self.hostname, self.port)

        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(self.timeout)
            sock.connect((self.hostname, self.port))

        except (socket.error, socket.timeout, ConnectionError):
            raise OnboardException(
                reason="fail-connect", message=f"ERROR device unreachable: {self.hostname}:{self.port}"
            )

    def guess_netmiko_device_type(self):
        """Guess the device type of host, based on Netmiko."""
        guessed_device_type = None

        netmiko_optional_args = netmiko_args(self.optional_args)

        remote_device = {
            "device_type": "autodetect",
            "host": self.hostname,
            "username": self.username,
            "password": self.password,
            **netmiko_optional_args,
        }

        if self.secret:
            remote_device["secret"] = self.secret

        if self.port:
            remote_device["port"] = self.port

        if self.timeout:
            remote_device["timeout"] = self.timeout

        try:
            logger.info("INFO guessing device type: %s", self.hostname)
            guesser = SSHDetect(**remote_device)
            guessed_device_type = guesser.autodetect()
            logger.info("INFO guessed device type: %s", guessed_device_type)

        except NetMikoAuthenticationException as err:
            logger.error("ERROR %s", err)
            raise OnboardException(reason="fail-login", message=f"ERROR: {str(err)}")

        except (NetMikoTimeoutException, SSHException) as err:
            logger.error("ERROR: %s", str(err))
            raise OnboardException(reason="fail-connect", message=f"ERROR: {str(err)}")

        except Exception as err:
            logger.error("ERROR: %s", str(err))
            raise OnboardException(reason="fail-general", message=f"ERROR: {str(err)}")

        logger.info("INFO device type is: %s", guessed_device_type)

        return guessed_device_type

    def set_napalm_driver_name(self):
        """Sets napalm driver name."""
        if not self.napalm_driver:
            netmiko_device_type = self.guess_netmiko_device_type()
            logger.info("Guessed Netmiko Device Type: %s", netmiko_device_type)

            self.netmiko_device_type = netmiko_device_type

            platform_to_napalm_netbox = {
                platform.slug: platform.napalm_driver for platform in Platform.objects.all() if platform.napalm_driver
            }

            # Update Constants if Napalm driver is defined for NetBox Platform
            netmiko_to_napalm = {**NETMIKO_TO_NAPALM_STATIC, **platform_to_napalm_netbox}

            self.napalm_driver = netmiko_to_napalm.get(netmiko_device_type)

    def check_napalm_driver_name(self):
        """Checks for napalm driver name."""
        if not self.napalm_driver:
            raise OnboardException(
                reason="fail-general",
                message=f"Onboarding for Platform {self.netmiko_device_type} not "
                f"supported, as it has no specified NAPALM driver",
            )

    def get_onboarding_facts(self):
        """Gather information from the network device that is needed to onboard the device into the NetBox system.

        Raises:
          OnboardException('fail-login'):
            When unable to login to device

          OnboardException('fail-execute'):
            When unable to run commands to collect device information

          OnboardException('fail-general'):
            Any other unexpected device comms failure.
        """
        self.check_reachability()

        logger.info("COLLECT: device information %s", self.hostname)

        try:
            # Get Napalm Driver with Netmiko if needed
            self.set_napalm_driver_name()

            # Raise if no Napalm Driver not selected
            self.check_napalm_driver_name()

            driver = get_network_driver(self.napalm_driver)

            optional_args = self.optional_args.copy()
            if self.port:
                optional_args["port"] = self.port

            if self.secret:
                optional_args["secret"] = self.secret

            napalm_device = driver(
                hostname=self.hostname,
                username=self.username,
                password=self.password,
                timeout=self.timeout,
                optional_args=optional_args,
            )

            napalm_device.open()

            logger.info("COLLECT: device facts")
            self.facts = napalm_device.get_facts()

            logger.info("COLLECT: device interface IPs")
            self.ip_ifs = napalm_device.get_interfaces_ip()

            module_name = PLUGIN_SETTINGS["onboarding_extensions_map"].get(self.napalm_driver)

            if module_name and self.load_driver_extension:
                try:
                    module = importlib.import_module(module_name)
                    driver_addon_class = module.OnboardingDriverExtensions(napalm_device=napalm_device)
                    self.onboarding_class = driver_addon_class.onboarding_class
                    self.driver_addon_result = driver_addon_class.ext_result
                except ModuleNotFoundError as exc:
                    raise OnboardException(
                        reason="fail-general",
                        message=f"ERROR: ModuleNotFoundError: Onboarding extension for napalm driver {self.napalm_driver} configured but can not be imported per configuration",
                    )
                except ImportError as exc:
                    raise OnboardException(reason="fail-general", message="ERROR: ImportError: %s" % exc.args[0])
            else:
                logger.info(
                    "INFO: No onboarding extension defined for napalm driver %s, using default napalm driver",
                    self.napalm_driver,
                )

        except ConnectionException as exc:
            raise OnboardException(reason="fail-login", message=exc.args[0])

        except CommandErrorException as exc:
            raise OnboardException(reason="fail-execute", message=exc.args[0])

        except Exception as exc:
            raise OnboardException(reason="fail-general", message=str(exc))

    def get_netdev_dict(self):
        """Construct network device dict."""
        netdev_dict = {
            "netdev_hostname": self.facts["hostname"],
            "netdev_vendor": self.facts["vendor"].title(),
            "netdev_model": self.facts["model"].lower(),
            "netdev_serial_number": self.facts["serial_number"],
            "netdev_mgmt_ifname": get_mgmt_info(hostname=self.hostname, ip_ifs=self.ip_ifs)[0],
            "netdev_mgmt_pflen": get_mgmt_info(hostname=self.hostname, ip_ifs=self.ip_ifs)[1],
            "netdev_netmiko_device_type": self.netmiko_device_type,
            "onboarding_class": self.onboarding_class,
            "driver_addon_result": self.driver_addon_result,
        }

        return netdev_dict



"""Forms for network device onboarding.

(c) 2020 Network To Code
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from django import forms
from django.db import transaction
from django_rq import get_queue

from utilities.forms import BootstrapMixin, CSVModelForm
from dcim.models import Site, Platform, DeviceRole, DeviceType

from .models import OnboardingTask
from .choices import OnboardingStatusChoices, OnboardingFailChoices
from .utils.credentials import Credentials

BLANK_CHOICE = (("", "---------"),)


class OnboardingTaskForm(BootstrapMixin, forms.ModelForm):
    """Form for creating a new OnboardingTask instance."""

    ip_address = forms.CharField(
        required=True, label="IP address", help_text="IP Address/DNS Name of the device to onboard"
    )

    site = forms.ModelChoiceField(required=True, queryset=Site.objects.all())

    username = forms.CharField(required=False, help_text="Device username (will not be stored in database)")
    password = forms.CharField(
        required=False, widget=forms.PasswordInput, help_text="Device password (will not be stored in database)"
    )
    secret = forms.CharField(
        required=False, widget=forms.PasswordInput, help_text="Device secret (will not be stored in database)"
    )

    platform = forms.ModelChoiceField(
        queryset=Platform.objects.all(),
        required=False,
        to_field_name="slug",
        help_text="Device platform. Define ONLY to override auto-recognition of platform.",
    )
    role = forms.ModelChoiceField(
        queryset=DeviceRole.objects.all(),
        required=False,
        to_field_name="slug",
        help_text="Device role. Define ONLY to override auto-recognition of role.",
    )
    device_type = forms.ModelChoiceField(
        queryset=DeviceType.objects.all(),
        required=False,
        to_field_name="slug",
        help_text="Device type. Define ONLY to override auto-recognition of type.",
    )

    class Meta:  # noqa: D106 "Missing docstring in public nested class"
        model = OnboardingTask
        fields = [
            "site",
            "ip_address",
            "port",
            "timeout",
            "username",
            "password",
            "secret",
            "platform",
            "role",
            "device_type",
        ]

    def save(self, commit=True, **kwargs):
        """Save the model, and add it and the associated credentials to the onboarding worker queue."""
        model = super().save(commit=commit, **kwargs)
        if commit:
            credentials = Credentials(self.data.get("username"), self.data.get("password"), self.data.get("secret"))
            get_queue("default").enqueue("netbox_onboarding.worker.onboard_device", model.pk, credentials)
        return model


class OnboardingTaskFilterForm(BootstrapMixin, forms.ModelForm):
    """Form for filtering OnboardingTask instances."""

    site = forms.ModelChoiceField(queryset=Site.objects.all(), required=False, to_field_name="slug")

    platform = forms.ModelChoiceField(queryset=Platform.objects.all(), required=False, to_field_name="slug")

    status = forms.ChoiceField(choices=BLANK_CHOICE + OnboardingStatusChoices.CHOICES, required=False)

    failed_reason = forms.ChoiceField(
        choices=BLANK_CHOICE + OnboardingFailChoices.CHOICES, required=False, label="Failed Reason"
    )

    q = forms.CharField(required=False, label="Search")

    class Meta:  # noqa: D106 "Missing docstring in public nested class"
        model = OnboardingTask
        fields = ["q", "site", "platform", "status", "failed_reason"]


class OnboardingTaskFeedCSVForm(CSVModelForm):
    """Form for entering CSV to bulk-import OnboardingTask entries."""

    site = forms.ModelChoiceField(
        queryset=Site.objects.all(),
        required=True,
        to_field_name="slug",
        help_text="Slug of parent site",
        error_messages={"invalid_choice": "Site not found",},
    )
    ip_address = forms.CharField(required=True, help_text="IP Address of the onboarded device")
    username = forms.CharField(required=False, help_text="Username, will not be stored in database")
    password = forms.CharField(required=False, help_text="Password, will not be stored in database")
    secret = forms.CharField(required=False, help_text="Secret password, will not be stored in database")
    platform = forms.ModelChoiceField(
        queryset=Platform.objects.all(),
        required=False,
        to_field_name="slug",
        help_text="Slug of device platform. Define ONLY to override auto-recognition of platform.",
        error_messages={"invalid_choice": "Platform not found.",},
    )
    port = forms.IntegerField(required=False, help_text="Device PORT (def: 22)",)

    timeout = forms.IntegerField(required=False, help_text="Device Timeout (sec) (def: 30)",)

    role = forms.ModelChoiceField(
        queryset=DeviceRole.objects.all(),
        required=False,
        to_field_name="slug",
        help_text="Slug of device role. Define ONLY to override auto-recognition of role.",
        error_messages={"invalid_choice": "DeviceRole not found",},
    )

    device_type = forms.ModelChoiceField(
        queryset=DeviceType.objects.all(),
        required=False,
        to_field_name="slug",
        help_text="Slug of device type. Define ONLY to override auto-recognition of type.",
        error_messages={"invalid_choice": "DeviceType not found",},
    )

    class Meta:  # noqa: D106 "Missing docstring in public nested class"
        model = OnboardingTask
        fields = [
            "site",
            "ip_address",
            "port",
            "timeout",
            "platform",
            "role",
        ]

    def save(self, commit=True, **kwargs):
        """Save the model, and add it and the associated credentials to the onboarding worker queue."""
        model = super().save(commit=commit, **kwargs)
        if commit:
            credentials = Credentials(self.data.get("username"), self.data.get("password"), self.data.get("secret"))
            transaction.on_commit(
                lambda: get_queue("default").enqueue("netbox_onboarding.worker.onboard_device", model.pk, credentials)
            )
        return model



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['netbox_onboarding',
 'netbox_onboarding.api',
 'netbox_onboarding.migrations',
 'netbox_onboarding.onboarding',
 'netbox_onboarding.onboarding_extensions',
 'netbox_onboarding.tests',
 'netbox_onboarding.utils']

package_data = \
{'': ['*'], 'netbox_onboarding': ['templates/netbox_onboarding/*']}

install_requires = \
['napalm>=2.5.0,<4', 'zipp>=3.4.0,<4.0.0']

setup_kwargs = {
    'name': 'ntc-netbox-plugin-onboarding',
    'version': '2.1.0',
    'description': 'A plugin for NetBox to easily onboard new devices.',
    'long_description': '# NetBox Onboarding plugin\n\n<!-- Build status with linky to the builds for ease of access. -->\n[![Build Status](https://travis-ci.com/networktocode/ntc-netbox-plugin-onboarding.svg?token=29s5AiDXdkDPwzSmDpxg&branch=master)](https://travis-ci.com/networktocode/ntc-netbox-plugin-onboarding)\n\nA plugin for [NetBox](https://github.com/netbox-community/netbox) to easily onboard new devices.\n\n`ntc-netbox-plugin-onboarding` is using [Netmiko](https://github.com/ktbyers/netmiko), [NAPALM](https://napalm.readthedocs.io/en/latest/) & [Django-RQ](https://github.com/rq/django-rq) to simplify the onboarding process of a new device into NetBox down to an IP Address and a site.\nThe goal of this plugin is not to import everything about a device into NetBox but rather to help build quickly an inventory in NetBox that is often the first step into an automation journey.\n\n## Installation\n\nIf using the installation pattern from the NetBox Documentation, you will need to activate the\nvirtual environment before installing so that you install the package into the virtual environment.\n\n```shell\ncd /opt/netbox\nsource venv/bin/activate\n```\n\nThe plugin is available as a Python package in pypi and can be installed with pip. Once the\ninstallation is completed, then NetBox and the NetBox worker must be restarted.\n\n```shell\npip install ntc-netbox-plugin-onboarding\nsystemctl restart netbox netbox-rq\n```\n\n### Compatibility Matrix\n\n|                       | Netbox 2.8 | Netbox 2.9 | Netbox 2.10 |\n|-----------------------|------------|------------|-------------|\n| Onboarding Plugin 1.3 |      X     |            |             |\n| Onboarding Plugin 2.0 |      X     |      X     |             |\n| Onboarding Plugin 2.1 |      X     |      X     |      X      |\n\nTo ensure NetBox Onboarding plugin is automatically re-installed during future upgrades, create a file named `local_requirements.txt` (if not already existing) in the NetBox root directory (alongside `requirements.txt`) and list the `ntc-netbox-plugin-onboarding` package:\n\n```no-highlight\n# echo ntc-netbox-plugin-onboarding >> local_requirements.txt\n```\n\nOnce installed, the plugin needs to be enabled in your `configuration.py`\n```python\n# In your configuration.py\nPLUGINS = ["netbox_onboarding"]\n\n# PLUGINS_CONFIG = {\n#   "netbox_onboarding": {\n#     ADD YOUR SETTINGS HERE\n#   }\n# }\n```\n\nFinally, make sure to run the migrations for this plugin\n\n```bash\npython3 manage.py migrate\n```\n\nThe plugin behavior can be controlled with the following list of settings\n\n- `create_platform_if_missing` boolean (default True), If True, a new platform object will be created if the platform discovered by netmiko do not already exist and is in the list of supported platforms (`cisco_ios`, `cisco_nxos`, `arista_eos`, `juniper_junos`, `cisco_xr`)\n- `create_device_type_if_missing` boolean (default True), If True, a new device type object will be created if the model discovered by Napalm do not match an existing device type.\n- `create_manufacturer_if_missing` boolean (default True), If True, a new manufacturer object will be created if the manufacturer discovered by Napalm is do not match an existing manufacturer, this option is only valid if `create_device_type_if_missing` is True as well.\n- `create_device_role_if_missing` boolean (default True), If True, a new device role object will be created if the device role provided was not provided as part of the onboarding and if the `default_device_role` do not already exist.\n- `create_management_interface_if_missing` boolean (default True), If True, add management interface and IP address to the device. If False no management interfaces will be created, nor will the IP address be added to NetBox, while the device will still get added.\n- `default_device_status` string (default "active"), status assigned to a new device by default (must be lowercase).\n- `default_device_role` string (default "network")\n- `default_device_role_color` string (default FF0000), color assigned to the device role if it needs to be created.\n- `default_management_interface` string (default "PLACEHOLDER"), name of the management interface that will be created, if one can\'t be identified on the device.\n- `default_management_prefix_length` integer ( default 0), length of the prefix that will be used for the management IP address, if the IP can\'t be found.\n- `skip_device_type_on_update` boolean (default False), If True, an existing NetBox device will not get its device type updated. If False, device type will be updated with one discovered on a device.\n- `skip_manufacturer_on_update` boolean (default False), If True, an existing NetBox device will not get its manufacturer updated. If False, manufacturer will be updated with one discovered on a device.\n- `platform_map` (dictionary), mapping of an **auto-detected** Netmiko platform to the **NetBox slug** name of your Platform. The dictionary should be in the format:\n    ```python\n    {\n      <Netmiko Platform>: <NetBox Slug>\n    }\n    ```\n- `onboarding_extensions_map` (dictionary), mapping of a NAPALM driver name to the loadable Python module used as an onboarding extension. The dictionary should be in the format:\n    ```python\n    {\n      <Napalm Driver Name>: <Loadable Python Module>\n    }\n    ```\n- `object_match_strategy` (string), defines the method for searching models. There are\ncurrently two strategies, strict and loose. Strict has to be a direct match, normally \nusing a slug. Loose allows a range of search criteria to match a single object. If multiple\nobjects are returned an error is raised. \n\n## Upgrades\n\nWhen a new release comes out it may be necessary to run a migration of the database to account for any changes in the data models used by this plugin. Execute the command `python3 manage.py migrate`from the NetBox install `netbox/` directory after updating the package.\n\n## Usage\n\n### Preparation\n\nTo properly onboard a device, the plugin needs to only know the Site as well as device\'s primary IP address or DNS Name.\n\n> For DNS Name Resolution to work, the instance of NetBox must be able to resolve the name of the\n> device to IP address.\n\nProviding other attributes (`Platform`, `Device Type`, `Device Role`) is optional - if any of these attributes is provided, plugin will use provided value for the onboarded device.\nIf `Platform`, `Device Type` and/or `Device Role` are not provided, the plugin will try to identify these information automatically and, based on the settings, it can create them in NetBox as needed.\n> If the Platform is provided, it must point to an existing NetBox Platform. NAPALM driver of this platform will be used only if it is defined for the platform in NetBox.\n> To use a preferred NAPALM driver, either define it in NetBox per platform or in the plugins settings under `platform_map`\n\n### Onboard a new device\n\nA new device can be onboarded via :\n- A web form  `/plugins/onboarding/add/`\n- A CSV form to import multiple devices in bulk. `/plugins/onboarding/import/`\n- An API, `POST /api/plugins\u200b/onboarding\u200b/onboarding\u200b/`\n\nDuring a successful onboarding process, a new device will be created in NetBox with its management interface and its primary IP assigned. The management interface will be discovered on the device based on the IP address provided.\n\n> By default, the plugin is using the credentials defined in the main `configuration.py` for Napalm (`NAPALM_USERNAME`/`NAPALM_PASSWORD`). It\'s possible to define specific credentials for each onboarding task.\n\n### Consult the status of onboarding tasks\n\nThe status of the onboarding process for each device is maintained is a dedicated table in NetBox and can be retrived :\n- Via the UI `/plugins/onboarding/`\n- Via the API `GET /api/plugins\u200b/onboarding\u200b/onboarding\u200b/`\n\n### API\n\nThe plugin includes 4 API endpoints to manage the onbarding tasks\n\n```shell\nGET        /api/plugins\u200b/onboarding\u200b/onboarding\u200b/       Check status of all onboarding tasks.\nPOST    \u200b   /api/plugins\u200b/onboarding\u200b/onboarding\u200b/       Onboard a new device\nGET     \u200b   /api/plugins\u200b/onboarding\u200b/onboarding\u200b/{id}\u200b/  Check the status of a specific onboarding task\nDELETE    \u200b /api/plugins\u200b/onboarding\u200b/onboarding\u200b/{id}\u200b/  Delete a specific onboarding task\n```\n\n## Contributing\n\nPull requests are welcomed and automatically built and tested against multiple version of Python and multiple version of NetBox through TravisCI.\n\nThe project is packaged with a light development environment based on `docker-compose` to help with the local development of the project and to run the tests within TravisCI.\n\nThe project is following Network to Code software development guideline and is leveraging:\n- Black, Pylint, Bandit and pydocstyle for Python linting and formatting.\n- Django unit test to ensure the plugin is working properly.\n\n### CLI Helper Commands\n\nThe project is coming with a CLI helper based on [invoke](http://www.pyinvoke.org/) to help setup the development environment. The commands are listed below in 3 categories `dev environment`, `utility` and `testing`. \n\nEach command can be executed with `invoke <command>`. All commands support the arguments `--netbox-ver` and `--python-ver` if you want to manually define the version of Python and NetBox to use. Each command also has its own help `invoke <command> --help`\n\n#### Local dev environment\n```\n  build            Build all docker images.\n  debug            Start NetBox and its dependencies in debug mode.\n  destroy          Destroy all containers and volumes.\n  start            Start NetBox and its dependencies in detached mode.\n  stop             Stop NetBox and its dependencies.\n```\n\n#### Utility \n```\n  cli              Launch a bash shell inside the running NetBox container.\n  create-user      Create a new user in django (default: admin), will prompt for password.\n  makemigrations   Run Make Migration in Django.\n  nbshell          Launch a nbshell session.\n```\n#### Testing \n\n```\n  tests            Run all tests for this plugin.\n  pylint           Run pylint code analysis.\n  pydocstyle       Run pydocstyle to validate docstring formatting adheres to NTC defined standards.\n  bandit           Run bandit to validate basic static code security analysis.\n  black            Run black to check that Python files adhere to its style standards.\n  unittest         Run Django unit tests for the plugin.\n```\n\n## Questions\n\nFor any questions or comments, please check the [FAQ](FAQ.md) first and feel free to swing by the [Network to Code slack channel](https://networktocode.slack.com/) (channel #networktocode).\nSign up [here](http://slack.networktocode.com/)\n\n## Screenshots\n\nList of Onboarding Tasks\n![Onboarding Tasks](docs/images/onboarding_tasks_view.png)\n\nCSV form to import multiple devices\n![CSV Form](docs/images/csv_import_view.png)\n\nOnboard a single device\n![Single Device Form](docs/images/single_device_form.png)\n\nMenu \n![Menu](docs/images/menu.png)\n\n',
    'author': 'Network to Code, LLC',
    'author_email': 'info@networktocode.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/networktocode/ntc-netbox-plugin-onboarding',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
